--[[---------------------------------------------------------------------------------
 AceLibrary by ckknight <ckknight@gmail.com>
           and Cladhaire <cladhaire@gmail.com>
 Based on the concept by Iriel <iriel@vigilance-committee.org>
 Design influenced by Tekkub <tekkub@gmail.com>

  This aims to be a standard for Ace embedded libraries, as the general stub they
  all use to register themselves.  This concept of embedded libraries came as a
  result of a number of discussions in both #wowace and #wowi-lounge.

  See http://www.iriel.org/wow/addondev/embedlibrary1.html for more information
----------------------------------------------------------------------------------]]

-- DO NOT CHANGE THESE VALUES -----------------------------
local ACELIBRARY_MAJOR = "AceLibrary"
local ACELIBRARY_MINOR = "$Revision: 3344 $"
-- DO NOT CHANGE THESE VALUES -----------------------------

-- CHANGE DEBUG TO ``false`` ON RELEASE -------------------
local DEBUG = true
-- CHANGE DEBUG TO ``false`` ON RELEASE -------------------

local _G = getfenv(0)
local previous = _G[ACELIBRARY_MAJOR]
if previous and not previous:IsNewVersion(ACELIBRARY_MAJOR, ACELIBRARY_MINOR) then return end

-- @table AceLibrary
-- @brief System to handle all versioning of libraries.
local AceLibrary = {}
local AceLibrary_mt = {}
setmetatable(AceLibrary, AceLibrary_mt)

local recurse = {}
local function addToPositions(t, major)
	if not AceLibrary.positions[t] or AceLibrary.positions[t] == major then
		t[recurse] = true
		AceLibrary.positions[t] = major
		for k,v in pairs(t) do
			if type(v) == "table" and not v[recurse] then
				addToPositions(v, major)
			end
			if type(k) == "table" and not k[recurse] then
				addToPositions(k, major)
			end
		end
		local mt = getmetatable(t)
		if mt and not mt[recurse] then
			addToPositions(mt, major)
		end
		t[recurse] = nil
	end
end

local function svnRevisionToNumber(text)
	if type(text) == "string" then
		if string.find(text, "^%$Revision: (%d+) %$$") then
			return tonumber((string.gsub(text, "^%$Revision: (%d+) %$$", "%1")))
		elseif string.find(text, "^%$Rev: (%d+) %$$") then
			return tonumber((string.gsub(text, "^%$Rev: (%d+) %$$", "%1")))
		elseif string.find(text, "^%$LastChangedRevision: (%d+) %$$") then
			return tonumber((string.gsub(text, "^%$LastChangedRevision: (%d+) %$$", "%1")))
		end
	elseif type(text) == "number" then
		return text
	end
	return nil
end

local crawlReplace
do
	local recurse = {}
	function crawlReplace(t, to, from)
		t[from], t[to] = nil, t[from]
		t[recurse] = true
		for k,v in pairs(t) do
			if k ~= recurse then
				if v == from then
					t[k] = to
				elseif type(v) == "table" then
					if not v[recurse] then
						crawlReplace(v, to, from)
					end
				end
				
				if type(k) == "table" then
					if not k[recurse] then
						crawlReplace(k, to, from)
					end
				end
			end
		end
		local mt = getmetatable(t)
		if mt then
			if mt == from then
				setmetatable(t, to)
			else
				crawlReplace(mt, to, from)
			end
		end
		t[recurse] = nil
	end
end

-- @function destroyTable
-- @brief    remove all the contents of a table
-- @param t  table to destroy
local function destroyTable(t)
	setmetatable(t, nil)
	for k,v in pairs(t) do t[k] = nil end
	table.setn(t, 0)
end

local new, del
do
	local tables = setmetatable({}, {__mode = "k"})
	
	function new()
		local t = next(tables)
		if t then
			tables[t] = nil
			return t
		else
			return {}
		end
	end
	
	function del(t, depth)
		if depth and depth > 0 then
			for k,v in pairs(t) do
				if type(v) == "table" then
					del(v, depth - 1)
				end
			end
		end
		destroyTable(t)
		tables[t] = true
	end
end

-- @function   copyTable
-- @brief      Create a shallow copy of a table and return it.
-- @param from The table to copy from
-- @return     A shallow copy of the table
local function copyTable(from)
	local to = new()
	for k,v in pairs(from) do to[k]=v end
	table.setn(to, table.getn(from))
	setmetatable(to, getmetatable(from))
	return to
end

-- @function         deepTransfer
-- @brief            Fully transfer all data, keeping proper previous table
--                   backreferences stable.
-- @param to         The table with which data is to be injected into
-- @param from       The table whose data will be injected into the first
-- @param saveFields If available, a shallow copy of the basic data is saved
--                   in here.
-- @param list       The account of table references
-- @param list2      The current status on which tables have been traversed.
local deepTransfer
do
	-- @function   examine
	-- @brief      Take account of all the table references to be shared
	--             between the to and from tables.
	-- @param to   The table with which data is to be injected into
	-- @param from The table whose data will be injected into the first
	-- @param list An account of the table references
	local function examine(to, from, list, major)
		list[from] = to
		for k in pairs(from) do
			if to[k] and type(from[k]) == "table" and type(to[k]) == "table" and not list[from[k]] then
				if from[k] == to[k] then
					list[from[k]] = to[k]
				elseif AceLibrary.positions[from[v]] ~= major and AceLibrary.positions[from[v]] then
					list[from[k]] = from[k]
				elseif not list[from[k]] then
					examine(to[k], from[k], list, major)
				end
			end
		end
		return list
	end
	
	function deepTransfer(to, from, saveFields, major, list, list2)
		setmetatable(to, nil)
		local createdList
		if not list then
			createdList = true
			list = new()
			list2 = new()
			examine(to, from, list, major)
		end
		list2[to] = to
		for k,v in pairs(to) do
			if not v or type(v) ~= "table" or type(v) ~= "table" then
				if saveFields then
					saveFields[k] = v
				end
				to[k] = nil
			elseif v ~= _G then
				if saveFields then
					saveFields[k] = copyTable(v)
				end
			end
		end
		for k in pairs(from) do
			if to[k] and to[k] ~= from[k] and AceLibrary.positions[to[k]] == major and from[k] ~= _G then
				if not list2[to[k]] then
					deepTransfer(to[k], from[k], nil, major, list, list2)
				end
				to[k] = list[to[k]] or list2[to[k]]
			else
				to[k] = from[k]
			end
		end
		table.setn(to, table.getn(from))
		setmetatable(to, getmetatable(from))
		local mt = getmetatable(to)
		if mt then
			if list[mt] then
				setmetatable(to, list[mt])
			elseif mt.__index and list[mt.__index] then
				mt.__index = list[mt.__index]
			end
		end
		destroyTable(from)
		if createdList then
			del(list)
			del(list2)
		end
	end
end

-- @method      IsNewVersion
-- @brief       Obtain whether the supplied version would be an upgrade to the
--              current version. This allows for bypass code in library
--              declaration.
-- @param major A string representing the major version
-- @param minor An integer or an svn revision string representing the minor version
-- @return      whether the supplied version would be newer than what is
--              currently available.
function AceLibrary:IsNewVersion(major, minor)
	if type(major) ~= "string" then
		error(string.format("Bad argument #2 to `IsNewVersion' (string expected, got %s)", type(major)), 2)
		return
	end
	if type(minor) == "string" then
		local m = svnRevisionToNumber(minor)
		if m then
			minor = m
		end
	end
	if type(minor) ~= "number" then
		error(string.format("Bad argument #3 to `IsNewVersion' (number expected, got %s)", type(minor)), 2)
		return
	end
	local data = self.libs[major]
	if not data then
		return true
	end
	return data.minor < minor
end

-- @method      HasInstance
-- @brief       Returns whether an instance exists. This allows for optional support of a library.
-- @param major A string representing the major version.
-- @param minor (optional) An integer or an svn revision string representing the minor version.
-- @return      Whether an instance exists.
function AceLibrary:HasInstance(major, minor)
	if type(major) ~= "string" then
		error(string.format("Bad argument #2 to `HasInstance' (string expected, got %s)", type(major)), 2)
		return
	end
	if minor then
		if type(minor) == "string" then
			local m = svnRevisionToNumber(minor)
			if m then
				minor = m
			end
		end
		if type(minor) ~= "number" then
			error(string.format("Bad argument #3 to `HasInstance' (number or nil expected, got %s)", type(minor)), 2)
			return
		end
		if not self.libs[major] then
			return
		end
		return self.libs[major].minor == minor
	end
	return self.libs[major] and true
end

-- @method      GetInstance
-- @brief       Returns the library with the given major/minor version.
-- @param major A string representing the major version.
-- @param minor (optional) An integer or an svn revision string representing the minor version.
-- @return      The library with the given major/minor version.
function AceLibrary:GetInstance(major, minor)
	if type(major) ~= "string" then
		error(string.format("Bad argument #2 to `GetInstance' (string expected, got %s)", type(major)), 2)
		return
	end

	local data = self.libs[major]
	if not data then
		error(string.format("Cannot find a library instance of %s.", major), 2)
		return
	end
	if minor then
		if type(minor) == "string" then
			local m = svnRevisionToNumber(minor)
			if m then
				minor = m
			end
		end
		if type(minor) ~= "number" then
			error(string.format("Bad argument #3 to `GetInstance' (number or nil expected, got %s)", type(minor)), 2)
			return
		elseif data.minor ~= minor then
			error(string.format("Cannot find a library instance of %s, minor version %d.", major, minor), 2)
			return
		end
	end
	return data.instance
end

-- Syntax sugar.  AceLibrary("FooBar-1.0")
AceLibrary_mt.__call = AceLibrary.GetInstance

local donothing

-- @method               Register
-- @brief                Registers a new version of a given library.
-- @param newInstance    the library to register
-- @param major          the major version of the library
-- @param minor          the minor version of the library
-- @param activateFunc   (optional) A function to be called when the library is
--                       fully activated. Takes the arguments
--                       (newInstance [, oldInstance, oldDeactivateFunc]). If
--                       oldInstance is given, you should probably call
--                       oldDeactivateFunc(oldInstance).
-- @param deactivateFunc (optional) A function to be called by a newer library's
--                       activateFunc.
function AceLibrary:Register(newInstance, major, minor, activateFunc, deactivateFunc)
	if type(newInstance) ~= "table" then
		error(string.format("Bad argument #2 to `Register' (table expected, got %s)", type(newInstance)), 2)
		return
	elseif type(major) ~= "string" then
		error(string.format("Bad argument #3 to `Register' (string expected, got %s)", type(major)), 2)
		return
	end
	if type(minor) == "string" then
		local m = svnRevisionToNumber(minor)
		if m then
			minor = m
		end
	end
	if type(minor) ~= "number" then
		error(string.format("Bad argument #4 to `Register' (number expected, got %s)", type(minor)), 2)
		return
	elseif math.floor(minor) ~= minor or minor < 0 then
		error(string.format("Bad argument #4 to `Register' (integer >= 0 expected, got %s)", minor), 2)
		return
	elseif activateFunc ~= nil and type(activateFunc) ~= "function" then
		error(string.format("Bad argument #5 to `Register' (nil or function expected, got %s)", type(activateFunc)), 2)
		return
	elseif deactivateFunc ~= nil and type(deactivateFunc) ~= "function" then
		error(string.format("Bad argument #6 to `Register' (nil or function expected, got %s)", type(deactivateFunc)), 2)
		return
	end
	if not deactivateFunc then
		if not donothing then
			donothing = function() end
		end
		deactivateFunc = donothing
	end
	local data = self.libs[major]
	if not data then
		-- This is new
		local instance = copyTable(newInstance)
		crawlReplace(instance, instance, newInstance)
		destroyTable(newInstance)
		if self == newInstance then
			self = instance
			AceLibrary = instance
		end
		self.libs[major] = {
			instance = instance,
			minor = minor,
			deactivateFunc = deactivateFunc
		}
		function instance:GetLibraryVersion()
			return major, minor
		end
		addToPositions(instance, major)
		if activateFunc then
			activateFunc(instance, nil, nil) -- no old version, so explicit nil
		end
		return instance
	end
	local instance = data.instance
	if minor <= data.minor then
		if DEBUG then
			-- This one is already obsolete, raise an error.
			error(string.format("Obsolete library registered. %s is already registered at version %d. You are trying to register version %d. Hint: if not AceLibrary:IsNewVersion(%q, %d) then return end", major, data.minor, minor, major, minor), 2)
			return
		end
		return instance
	end
	
	-- This is an update
	local oldInstance = new()
	addToPositions(newInstance, major)
	deepTransfer(instance, newInstance, oldInstance, major)
	crawlReplace(instance, instance, newInstance)
	local oldDeactivateFunc = data.deactivateFunc
	data.minor = minor
	data.deactivateFunc = deactivateFunc
	function instance:GetLibraryVersion()
		return major, minor
	end
	if activateFunc then
		activateFunc(instance, oldInstance, oldDeactivateFunc)
	else
		oldDeactivateFunc(oldInstance)
	end
	del(oldInstance, 1)
	return instance
end

-- @function            Activate
-- @brief               The activateFunc for AceLibrary itself. Called when
--                      AceLibrary properly registers.
-- @param self          Reference to AceLibrary
-- @param oldLib        (optional) Reference to an old version of AceLibrary
-- @param oldDeactivate (optional) Function to deactivate the old lib
local function activate(self, oldLib, oldDeactivate)
	if not self.libs then
		if oldLib then
			self.libs = oldLib.libs
		else
			self.libs = {}
		end
	end
	if not self.positions then
		if oldLib then
			self.positions = oldLib.positions
		else
			self.positions = setmetatable({}, { __mode = "k" })
		end
	end
	
	-- Expose the library in the global environment
	_G[ACELIBRARY_MAJOR] = self
	
	if oldLib then
		oldDeactivate(oldLib)
	end
end

if previous then
	previous:Register(AceLibrary, ACELIBRARY_MAJOR, ACELIBRARY_MINOR, activate)
else
	AceLibrary.libs = {}
	AceLibrary.positions = setmetatable({}, { __mode = "k" })
	AceLibrary:Register(AceLibrary, ACELIBRARY_MAJOR, ACELIBRARY_MINOR, activate)
end
AceLibrary = AceLibrary(ACELIBRARY_MAJOR)