#!/usr/bin/env lua5.0

local MAJOR_VERSION = "AceOO-2.0"
local MINOR_VERSION = "$Revision: 3344 $"

-- This ensures the code is only executed if the libary doesn't already exist, or is a newer version
if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

-- @function	getuid
-- @brief		Obtain a unique string identifier for the object in question.
-- @param t		The object to obtain the uid for.
-- @return		The uid string.
local function pad(cap)
	return string.rep('0', 8 - string.len(cap)) .. cap
end
local function getuid(t)
	local str = tostring(t)
	local _,_,cap = string.find(str, '[^:]*: 0x(.*)$')
	if cap then return pad(cap) end
	_,_,cap = string.find(str, '[^:]*: (.*)$')
	if cap then return pad(cap) end
end

local function getlibrary(o)
	if type(o) == "table" then
		return o
	elseif type(o) == "string" then
		if not AceLibrary:HasInstance(o) then
			error(string.format("Library %q does not exist.", o), 3)
		end
		return AceLibrary(o)
	end
end

-- @function		Factory
-- @brief			Construct a factory for the creation of objects.
-- @param obj		The object whose init method will be called on the new factory
--					object.
-- @param newobj	The object whose init method will be called on the new
--					objects that the Factory creates, to initialize them.
-- @param (a1..a20) Arguments which will be passed to obj.init() in addition
--					to the Factory object.
-- @return			The new factory which creates a newobj when its new method is called,
--					or when it is called directly (__call metamethod).
local Factory
do
	local function new(obj, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12,
							a13, a14, a15, a16, a17, a18, a19, a20)
		local t = {}
		local uid = getuid(t)
		local l = getlibrary
		obj:init(t, l(a1), l(a2), l(a3), l(a4), l(a5), l(a6), l(a7),
					l(a8), l(a9), l(a10), l(a11), l(a12), l(a13),
					l(a14), l(a15), l(a16), l(a17), l(a18), l(a19),
					l(a20))
		t.uid = uid
		return t
	end
	
	local function createnew(self, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10,
									a11, a12, a13, a14, a15, a16, a17, a18,
									a19, a20)
		local o = self.prototype
		local l = getlibrary
		return new(o, l(a1), l(a2), l(a3), l(a4), l(a5), l(a6), l(a7),
						l(a8), l(a9), l(a10), l(a11), l(a12), l(a13),
						l(a14), l(a15), l(a16), l(a17), l(a18), l(a19),
						l(a20))
	end

	function Factory(obj, newobj, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10,
									a11, a12, a13, a14, a15, a16, a17, a18,
									a19, a20)
		local t = new(obj, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12,
							a13, a14, a15, a16, a17, a18, a19, a20)
		t.prototype = newobj
		t.new = createnew
		getmetatable(t).__call = t.new
		return t
	end
end


local function objtostring(self)
	if rawget(self, 'ToString') then
		return self:ToString()
	elseif self.ToString then
		local s = self:ToString() .. "("
		local first = true
		if self.interfaces then
			for interface in pairs(self.interfaces) do
				if first then
					s = s .. tostring(interface)
					first = false
				else
					s = s .. ", " .. tostring(interface)
				end
			end
		end
		if self.mixins then
			for mixin in pairs(self.mixins) do
				if first then
					s = s .. tostring(mixin)
					first = false
				else
					s = s .. ", " .. tostring(mixin)
				end
			end
		end
		return s .. ")"
	else
		return self.uid or 'table'
	end
end

-- @table			Object
-- @brief			Base of all objects, including Class.
--
-- @method			init
-- @brief			Initialize a new object.
-- @param newobject The object to initialize
-- @param class		The class to make newobject inherit from
local Object
do
	Object = {}
	function Object:init(newobject, class)
		local parent = class or self
		if not rawget(newobject, 'uid') then
			newobject.uid = getuid(newobject)
		end
		local mt = {
			__index = parent,
			__tostring = objtostring,
		}
		setmetatable(newobject, mt)
	end
	Object.uid = getuid(Object)
	setmetatable(Object, { __tostring = function() return 'Object' end })
end

local Interface

local function validateInterface(object, interface)
	if not object.class and object.prototype then
		object = object.prototype
	end
	for k,v in pairs(interface.interface) do
		if tostring(type(object[k])) ~= v then
			return false
		end
	end
	if interface.superinterfaces then
		for superinterface in pairs(interface.superinterfaces) do
			if not validateInterface(object, superinterface) then
				return false
			end
		end
	end
	if type(object.class) == "table" and rawequal(object.class.prototype, object) then
		if not object.class.interfaces then
			object.class.interfaces = {}
		end
		object.class.interfaces[interface] = true
	elseif type(object.class) == "table" and type(object.class.prototype) == "table" then
		validateInterface(object.class.prototype, interface)
		-- check if class is proper, thus preventing future checks.
	end
	return true
end

-- @function		inherits
-- @brief			Return whether an Object or Class inherits from a given
--					parent.
-- @param object	Object or Class to check
-- @param parent	Parent to test inheritance from
-- @return			whether an Object or Class inherits from a given
--					parent.
local function inherits(object, parent)
	object = getlibrary(object)
	if type(parent) == "string" then
		if not AceLibrary:HasInstance(parent) then
			return false
		else
			parent = AceLibrary(parent)
		end
	end
	if type(parent) ~= "table" then
		error(string.format("Bad argument #2 to `inherits' (table expected, got %s)", type(class)), 2)
	elseif type(object) ~= "table" then
		return false
	end
	local current
	if object.class then
		current = object.class
	else
		current = object
	end
	if type(current) ~= "table" then
		return false
	end
	if rawequal(current, parent) then
		return true
	end
	if parent.class then
		while true do
			if rawequal(current, Object) then
				break
			end
			if current.mixins then
				for mixin in pairs(current.mixins) do
					if rawequal(mixin, parent) then
						return true
					end
				end
			end
			if current.interfaces then
				for interface in pairs(current.interfaces) do
					if rawequal(interface, parent) then
						return true
					end
				end
			end
			current = current.super
			if type(current) ~= "table" then
				break
			end
		end
		
		local isInterface = false
		local curr = parent.class
		while true do
			if rawequal(curr, Object) then
				break
			elseif rawequal(curr, Interface) then
				isInterface = true
				break
			end
			curr = curr.super
			if type(curr) ~= "table" then
				break
			end
		end
		return isInterface and validateInterface(object, parent)
	else
		while true do
			if rawequal(current, parent) then
				return true
			elseif rawequal(current, Object) then
				return false
			end
			current = current.super
			if type(current) ~= "table" then
				return false
			end
		end
	end
end

-- @table			Class
-- @brief			An object factory which sets up inheritence and supports
--					'mixins'.
--
-- @metamethod		Class call
-- @brief			Call ClassFactory:new() to create a new class.
--
-- @method			Class new
-- @brief			Construct a new object.
-- @param (a1..a20) Arguments to pass to the object init function.
-- @return			The new object.
--
-- @method			Class init
-- @brief			Initialize a new class.
-- @param parent	Superclass.
-- @param (a1..a20) Mixins.
--
-- @method			Class ToString
-- @return			A string representing the object, in this case 'Class'.
local initStatus
local Class
local Mixin
local function traverseInterfaces(bit, total)
	if bit.superinterfaces then
		for interface in pairs(bit.superinterfaces) do
			if not total[interface] then
				total[interface] = true
				traverseInterfaces(interface, total)
			end
		end
	end
end
do
	Class = Factory(Object, setmetatable({}, {__index = Object}), Object)
	Class.super = Object
	Class.virtual = true
	
	local function protostring(t)
		if t.ToString then
			return t:ToString()
		else
			return '<' .. tostring(t.class) .. ' prototype>'
		end
	end
	local function classobjectstring(t)
		if t.ToString then
			return t:ToString()
		elseif t.GetLibraryVersion then
			return (t:GetLibraryVersion())
		else
			return '<' .. tostring(t.class) .. ' instance>'
		end
	end
	local function classobjectequal(self, other)
		if self.Equals then
			return self:Equals(other)
		elseif other.Equals then
			return other:Equals(self)
		else
			return rawequal(self, other)
		end
	end
	local function classobjectlessthan(self, other)
		if self.IsLessThan then
			return self:IsLessThan(other)
		elseif other.IsLessThanOrEqual then
			return not other:IsLessThanOrEqual(self)
		elseif self.CompareTo then
			return self:CompareTo(other) < 0
		elseif other.CompareTo then
			return other:CompareTo(self) > 0
		else
			error("cannot compare two objects", 2)
		end
	end
	local function classobjectlessthanequal(self, other)
		if self.IsLessThanOrEqual then
			return self:IsLessThanOrEqual(other)
		elseif other.IsLessThan then
			return not other:IsLessThan(self)
		elseif self.CompareTo then
			return self:CompareTo(other) <= 0
		elseif other.CompareTo then
			return other:CompareTo(self) >= 0
		else
			error("cannot compare two incompatible objects", 2)
		end
	end
	local function classobjectadd(self, other)
		if self.Add then
			return self:Add(other)
		else
			error("cannot add two incompatible objects", 2)
		end
	end
	local function classobjectsub(self, other)
		if self.Subtract then
			return self:Subtract(other)
		else
			error("cannot subtract two incompatible objects", 2)
		end
	end
	local function classobjectunm(self, other)
		if self.UnaryNegation then
			return self:UnaryNegation(other)
		else
			error("attempt to negate an incompatible object", 2)
		end
	end
	local function classobjectmul(self, other)
		if self.Multiply then
			return self:Multiply(other)
		else
			error("cannot multiply two incompatible objects", 2)
		end
	end
	local function classobjectdiv(self, other)
		if self.Divide then
			return self:Divide(other)
		else
			error("cannot divide two incompatible objects", 2)
		end
	end
	local function classobjectpow(self, other)
		if self.Exponent then
			return self:Exponent(other)
		else
			error("cannot exponentiate two incompatible objects", 2)
		end
	end
	local function classobjectconcat(self, other)
		if self.Concatenate then
			return self:Concatenate(other)
		else
			error("cannot concatenate two incompatible objects", 2)
		end
	end
	function Class:new(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12,
						a13, a14, a15, a16, a17, a18, a19, a20)
		if self.virtual then
			error("Cannot instantiate a virtual class.", 2)
		end
		
		local o = self.prototype
		local newobj = {}
		if o.class and o.class.instancemeta then
			setmetatable(newobj, o.class.instancemeta)
		else
			Object:init(newobj, o)
		end
		
		if self.interfaces and not self.interfacesVerified then
			-- Verify the interfaces
			
			for interface in pairs(self.interfaces) do
				for field,kind in pairs(interface.interface) do
					if tostring(type(newobj[field])) ~= kind then
						error(string.format("Class did not satisfy all interfaces. %q is required to be a %s. It is a %s", field, kind, tostring(type(newobj[field]))), 5)
					end
				end
			end
			self.interfacesVerified = true
		end
		local tmp = initStatus
		initStatus = newobj
		newobj:init(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12,
						a13, a14, a15, a16, a17, a18, a19, a20)
		if initStatus then
			initStatus = tmp
			error("Initialization not completed, be sure to call the superclass's init method.", 2)
			return
		end
		initStatus = tmp
		return newobj
	end
	function Class:init(newclass, parent, a1, a2, a3, a4, a5, a6, a7, a8, a9,
											a10, a11, a12, a13, a14, a15, a16,
											a17, a18, a19, a20)
		parent = parent or self
		
		local total
		
		if parent.class then
			total = {
				parent, a1, a2, a3, a4, a5, a6, a7, a8, a9,
						a10, a11, a12, a13, a14, a15, a16,
						a17, a18, a19, a20
			}
			parent = self
		else
			total = {
				a1, a2, a3, a4, a5, a6, a7, a8, a9,
				a10, a11, a12, a13, a14, a15, a16,
				a17, a18, a19, a20
			}
		end
		if not inherits(parent, Class) then
			error("Classes must inherit from a proper class", 4)
		end
		if parent.sealed then
			error("Cannot inherit from a sealed class", 4)
		end
		for i,v in ipairs(total) do
			if inherits(v, Mixin) and v.class then
				if not newclass.mixins then
					newclass.mixins = {}
				end
				if newclass.mixins[v] then
					error("Cannot explicitly inherit from the same mixin twice", 4)
				end
				newclass.mixins[v] = true
			elseif inherits(v, Interface) and v.class then
				if not newclass.interfaces then
					newclass.interfaces = {}
				end
				if newclass.interfaces[v] then
					error("Cannot explicitly inherit from the same interface twice", 4)
				end
				newclass.interfaces[v] = true
			else
				error("Classes can only inherit from one or zero classes and any number of mixins or interfaces", 4)
			end
		end
		if parent.interfaces then
			if newclass.interfaces then
				for interface in pairs(parent.interfaces) do
					newclass.interfaces[interface] = true
				end
			else
				newclass.interfaces = parent.interfaces
			end
		end
		for k in pairs(total) do
			total[k] = nil
		end
		table.setn(total, 0)
		newclass.mixinmethods = setmetatable(total, {__index = parent.prototype})
		total = nil
		
		newclass.super = parent
		
		newclass.protometa = {
			__index = newclass.mixinmethods,
			__tostring = protostring,
		}
		newclass.prototype = {class = newclass}
		setmetatable(newclass.prototype, newclass.protometa)

		newclass.instancemeta = {
			__index = newclass.prototype,
			__tostring = classobjectstring,
			__eq = classobjectequal,
			__lt = classobjectlessthan,
			__le = classobjectlessthanequal,
			__add = classobjectadd,
			__sub = classobjectsub,
			__unm = classobjectunm,
			__mul = classobjectmul,
			__div = classobjectdiv,
			__pow = classobjectpow,
			__concat = classobjectconcat,
		}

		local classmeta = {
			__index = self,
			__tostring = objtostring,
			__call = self.new,
		}
		setmetatable(newclass, classmeta)
		
		if not rawget(newclass, 'virtual') then
			newclass.virtual = false
		end
		newclass.interfacesVerified = false
		
		if newclass.mixins then
			-- Fold in the mixins
			local err, msg
			for mixin in pairs(newclass.mixins) do
				local ret
				ret, msg = pcall(mixin.embed, mixin, newclass.mixinmethods)
				if not ret then
					err = true
					break
				end
			end
	
			if err then
				local mm = newclass.mixinmethods
				for k,v in pairs(mm) do
					mm[k] = nil
				end
	
				-- method conflict
				error(msg, 4)
			end
		end
		
		if newclass.interfaces then
			for interface in pairs(newclass.interfaces) do
				traverseInterfaces(interface, newclass.interfaces)
			end
		end
		if newclass.mixins then
			for mixin in pairs(newclass.mixins) do
				if mixin.interfaces then
					if not newclass.interfaces then
						newclass.interfaces = {}
					end
					for interface in pairs(mixin.interfaces) do
						newclass.interfaces[interface] = true
					end
				end
			end
		end
	end
	function Class:ToString()
		if type(self.GetLibraryVersion) == "function" then
			return (self:GetLibraryVersion())
		else
			return "Class"
		end
	end
	
	local tmp
	function Class.prototype:init()
		if rawequal(self, initStatus) then
			initStatus = nil
		end
		local current = self.class
		while true do
			if current == Class then
				break
			end
			if current.mixins then
				for mixin in pairs(current.mixins) do
					if type(mixin.OnInstanceInit) == "function" then
						mixin:OnInstanceInit(self)
					end
				end
			end
			current = current.super
		end
	end
end


-- @object	ClassFactory
-- @brief	A factory for creating classes.	Rarely used directly.
local ClassFactory = Factory(Object, Class, Object)

getmetatable(Class).__call =
	function (t, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11,
				 a12, a13, a14, a15, a16, a17, a18, a19, a20)
		return ClassFactory:new(a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11,
								a12, a13, a14, a15, a16, a17, a18, a19, a20)
	end

-- @class	Mixin
-- @brief	A class to create mixin objects, which contain methods that get
-- "mixed in" to class prototypes.
--
-- @object	Mixin prototype
-- @brief	The prototype that mixin objects inherit their methods from.
--
-- @method	Mixin prototype embed
-- @brief	Mix in the methods of our object which are listed in our interface
--		 to the supplied target table.
--
-- @method	Mixin prototype init
-- @brief	Initialize the mixin object.
-- @param	newobj	 The new object we're initializing.
-- @param	interface	The interface we implement (the list of methods our
--					prototype provides which should be mixed into the target
--					table by embed).
do
	Mixin = Class()
	function Mixin:ToString()
		if self.GetLibraryVersion then
			return (self:GetLibraryVersion())
		else
			return 'Mixin'
		end
	end
	local function _Embed(state, field, target)
		field = next(state.export, field)
		if field == nil then
			return
		end

		if rawget(target, field) or (target[field] and target[field] ~= state[field]) then
			error("Method conflict in attempt to mixin.", 2)
		end

		target[field] = state[field]

		local ret,msg = pcall(_Embed, state, field, target)
		if not ret then
			-- Mix in the next method according to the defined interface.	If that
			-- fails due to a conflict, re-raise to back out the previous mixed
			-- methods.

			target[field] = nil
			error(msg, 2)
		end
	end
	function Mixin.prototype:embed(target)
		local err, msg = pcall(_Embed, self, nil, target)
		if not err then
			error(msg, 2)
		end
		if type(self.embedList) == "table" then
			self.embedList[target] = true
		end
	end
	
	function Mixin.prototype:activate(oldLib, oldDeactivate)
		if oldLib and oldLib.embedList then
			for target in pairs(oldLib.embedList) do
				for field in pairs(oldLib.export) do
					target[field] = nil
				end
			end
			self.embedList = oldLib.embedList
			for target in pairs(self.embedList) do
				self:embed(target)
			end
		else
			self.embedList = setmetatable({}, {__mode="k"})
		end
	end
	
	function Mixin.prototype:init(export, a1, a2, a3, a4, a5, a6, a7, a8, a9,
											a10, a11, a12, a13, a14, a15, a16,
											a17, a18, a19, a20)
		if type(export) ~= "table" then
			error(string.format("Bad argument #2 to `new' (table expected, got %s)", type(export)), 3)
		end
		for k,v in pairs(export) do
			if type(k) ~= "number" then
				error("All keys to argument #2 must be numbers.", 3)
			elseif type(v) ~= "string" then
				error("All values to argument #2 must be strings.", 3)
			end
		end
		local num = table.getn(export)
		for i = 1, num do
			local v = export[i]
			export[i] = nil
			export[v] = true
		end
		table.setn(export, 0)
		local interfaces
		if a1 then
			local l = getlibrary
			interfaces = { l(a1), l(a2), l(a3), l(a4), l(a5), l(a6), l(a7),
				l(a8), l(a9), l(a10), l(a11), l(a12), l(a13), l(a14), l(a15),
				l(a16), l(a17), l(a18), l(a19), l(a20) }
			for _,v in ipairs(interfaces) do
				if not v.class or not inherits(v, Interface) then
					error("Mixins can inherit only from interfaces", 4)
				end
			end
			local num = table.getn(interfaces)
			for i = 1, num do
				local v = interfaces[i]
				interfaces[i] = nil
				interfaces[v] = true
			end
			table.setn(interfaces, 0)
			for interface in pairs(interfaces) do
				traverseInterfaces(interface, interfaces)
			end
			for interface in pairs(interfaces) do
				for field,kind in pairs(interface.interface) do
					if kind ~= "nil" then
						local good = false
						for bit in pairs(export) do
							if bit == field then
								good = true
								break
							end
						end
						if not good then
							error(string.format("Mixin does not fully accommodate field %q", field), 4)
						end
					end
				end
			end
		end
		self.super = Mixin.prototype
		Mixin.super.prototype.init(self)
		self.export = export
		self.interfaces = interfaces
	end
end

-- @class Interface
-- @brief A class to create interfaces, which contain contracts that classes
--			which inherit from this must comply with.
--
-- @object Interface prototype
-- @brief	The prototype that interface objects must adhere to.
--
-- @method Interface prototype init
-- @brief	Initialize the mixin object.
-- @param	interface	The interface we contract (the hash of fields forced).
-- @param	(a1..a20)	Superinterfaces
do
	Interface = Class()
	function Interface:ToString()
		if self.GetLibraryVersion then
			return (self:GetLibraryVersion())
		else
			return 'Instance'
		end
	end
	function Interface.prototype:init(interface, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, a17, a18, a19, a20)
		Interface.super.prototype.init(self)
		if type(interface) ~= "table" then
			error(string.format("Bad argument #2 to `new' (table expected, got %s)", type(interface)), 3)
		end
		for k,v in pairs(interface) do
			if type(k) ~= "string" then
				error("All keys to argument #2 must be numbers.", 3)
			elseif type(v) ~= "string" then
				error("All values to argument #2 must be strings.", 3)
			elseif v ~= "nil" and v ~= "string" and v ~= "number" and v ~= "table" and v ~= "function" then
				error('All values to argument #2 must either be "nil", "string", "number", "table", or "function".', 3)
			end
		end
		local l = getlibrary
		a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, a17, a18, a19, a20 = l(a1), l(a2), l(a3), l(a4), l(a5), l(a6), l(a7), l(a8), l(a9), l(a10), l(a11), l(a12), l(a13), l(a14), l(a15), l(a16), l(a17), l(a18), l(a19), l(a20)
		if a1 then
			self.superinterfaces = {a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, a17, a18, a19, a20}
			for k,v in ipairs(self.superinterfaces) do
				if not inherits(v, Interface) or not v.class then
					error('Cannot provide a non-Interface to inherit from', 3)
				end
			end
			local num = table.getn(self.superinterfaces)
			for i = 1, num do
				local v = self.superinterfaces[i]
				self.superinterfaces[i] = nil
				self.superinterfaces[v] = true
			end
			table.setn(self.superinterfaces, 0)
		end
		self.interface = interface
	end
end

-- @function Classpool
-- @brief	Obtain a read only class from our pool of classes, indexed by the
--		 superclass and mixins.
-- @param	sc		 The superclass of the class we want.
-- @param	(m1..m20)	Mixins of the class we want's objects.
-- @return A read only class from the class pool.
local Classpool
do
	local pool = setmetatable({}, {__mode = 'v'})
	local function newindex(k, v)
		error('Attempt to modify a read only class.', 2)
	end
	local t
	local function getcomplexuid(sc, m1, m2, m3, m4, m5, m6, m7, m8, m9,
		m10, m11, m12, m13, m14, m15, m16,
		m17, m18, m19, m20)
		if not t then t = {} end
		if sc then table.insert(t, sc.uid)
		if m1 then table.insert(t, m1.uid)
		if m2 then table.insert(t, m2.uid)
		if m3 then table.insert(t, m3.uid)
		if m4 then table.insert(t, m4.uid)
		if m5 then table.insert(t, m5.uid)
		if m6 then table.insert(t, m6.uid)
		if m7 then table.insert(t, m7.uid)
		if m8 then table.insert(t, m8.uid)
		if m9 then table.insert(t, m9.uid)
		if m10 then table.insert(t, m10.uid)
		if m11 then table.insert(t, m11.uid)
		if m12 then table.insert(t, m12.uid)
		if m13 then table.insert(t, m13.uid)
		if m14 then table.insert(t, m14.uid)
		if m15 then table.insert(t, m15.uid)
		if m16 then table.insert(t, m16.uid)
		if m17 then table.insert(t, m17.uid)
		if m18 then table.insert(t, m18.uid)
		if m19 then table.insert(t, m19.uid)
		if m20 then table.insert(t, m20.uid)
		end end end end end end end end end end end end end end end end end end end end end
		table.sort(t)
		local uid = table.concat(t, '')
		for k in pairs(t) do t[k] = nil end
		table.setn(t, 0)
		return uid
	end
	function Classpool(sc, m1, m2, m3, m4, m5, m6, m7, m8, m9,
		m10, m11, m12, m13, m14, m15, m16,
		m17, m18, m19, m20)
		local l = getlibrary
		sc, m1, m2, m3, m4, m5, m6, m7, m8, m9, m10, m11, m12, m13, m14, m15, m16, m17, m18, m19, m20 = l(sc), l(m1), l(m2), l(m3), l(m4), l(m5), l(m6), l(m7), l(m8), l(m9), l(m10), l(m11), l(m12), l(m13), l(m14), l(m15), l(m16), l(m17), l(m18), l(m19), l(m20)
		sc = sc or Class
		local key = getcomplexuid(sc, m1, m2, m3, m4, m5, m6, m7, m8, m9, m10, m11, m12, m13, m14, m15, m16, m17, m18, m19, m20)
		if not pool[key] then
			local class = Class(sc, m1, m2, m3, m4, m5, m6, m7, m8, m9,
			m10, m11, m12, m13, m14, m15, m16, m17,
			m18, m19, m20)
			-- Prevent the user from adding methods to this class.
			-- NOTE: I'm not preventing modifications of existing class members,
			-- but it's likely that only a truly malicious user will be doing so.
			getmetatable(class.prototype).__newindex = newindex
			getmetatable(class.prototype).__metatable = "Cannot manipulate the metatable of this object."
			class.sealed = true
			getmetatable(class).__newindex = newindex
			getmetatable(class).__metatable = "Cannot manipulate the metatable of this object."
			pool[key] = class
		end
		return pool[key]
	end
end

local obj = {
	Factory = Factory,
	Object = Object,
	Class = Class,
	Mixin = Mixin,
	Interface = Interface,
	Classpool = Classpool,
	inherits = inherits
}

-- Library handling bits

AceLibrary:Register(obj, MAJOR_VERSION, MINOR_VERSION)
